@@ -24,8 +24,6 @@
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)
 
-(* $Id: postgresql.ml,v 1.7 2006/01/24 21:11:09 mottl Exp $ *)
-
 open Printf
 
 type oid = int
@@ -232,8 +230,12 @@ let ftype_of_string = function
   | str -> failwith ("ftype_of_string: unknown ftype: " ^ str)
 
 external init : unit -> unit = "PQocaml_init"
+
+let null = ""
+
 let () =
   Callback.register_exception "Postgresql.Oid" (Oid invalid_oid);
+  Callback.register "Postgresql.null" null;
   init ()
 
 type connection_status = Ok | Bad
@@ -325,7 +327,13 @@ module Stub = struct
   (* Command Execution Functions *)
 
   external result_isnull : result -> bool = "PQres_isnull" "noalloc"
-  external exec : connection -> string -> result = "PQexec_stub"
+
+  external exec_params :
+    connection -> string -> string array -> bool array -> result
+    = "PQexecParams_stub"
+
+  external describe_prepared :
+    connection -> string -> result = "PQdescribePrepared_stub"
 
   external result_status :
     result -> result_status = "PQresultStatus_stub" "noalloc"
@@ -336,11 +344,21 @@ module Stub = struct
     connection -> result_status -> result = "PQmakeEmptyPGresult_stub"
 
   external ntuples : result -> int = "PQntuples_stub" "noalloc"
+
+(* FIXME: switch to noalloc once PostgreSQL 8.2 is out for CentOS *)
+(*   external nparams : result -> int = "PQnparams_stub" "noalloc" *)
+  external nparams : result -> int = "PQnparams_stub"
+
   external nfields : result -> int = "PQnfields_stub" "noalloc"
   external fname : result -> int -> string = "PQfname_stub"
   external fnumber : result -> string -> int ="PQfnumber_stub" "noalloc"
   external fformat : result -> int -> FFormat.t = "PQfformat_stub" "noalloc"
   external ftype : result -> int -> oid = "PQftype_stub" "noalloc"
+
+(* FIXME: switch to noalloc once PostgreSQL 8.2 is out for CentOS *)
+(*   external paramtype : result -> int -> oid = "PQparamtype_stub" "noalloc" *)
+  external paramtype : result -> int -> oid = "PQparamtype_stub"
+
   external fmod : result -> int -> int = "PQfmod_stub" "noalloc"
   external fsize : result -> int -> int = "PQfsize_stub" "noalloc"
   external binary_tuples : result -> bool = "PQbinaryTuples_stub" "noalloc"
@@ -366,8 +384,9 @@ module Stub = struct
   external is_nonblocking :
     connection -> bool = "PQisnonblocking_stub" "noalloc"
 
-  external send_query :
-    connection -> string -> int = "PQsendQuery_stub" "noalloc"
+  external send_query_params :
+    connection -> string -> string array -> bool array -> int
+    = "PQsendQueryParams_stub" "noalloc"
 
   external get_result : connection -> result = "PQgetResult_stub"
   external consume_input : connection -> int = "PQconsumeInput_stub" "noalloc"
@@ -451,10 +470,17 @@ external unescape_bytea : string -> string = "PQunescapeBytea_stub"
 class result res =
   let nfields = Stub.nfields res in
   let ntuples = Stub.ntuples res in
+(* FIXME: get rid of laziness once PostgreSQL 8.2 is out for CentOS *)
+(*   let nparams = Stub.nparams res in *)
+  let nparams = lazy (Stub.nparams res) in
   let binary_tuples = Stub.binary_tuples res in
   let check_field field =
     if field < 0 || field >= nfields then
       raise (Error (Field_out_of_range (field, nfields))) in
+  let check_param param =
+    let nparams = Lazy.force nparams in
+    if param < 0 || param >= nparams then
+      raise (Error (Field_out_of_range (param, nparams))) in
   let check_tuple tuple =
     if tuple < 0 || tuple >= ntuples then
       raise (Error (Tuple_out_of_range (tuple, ntuples))) in
@@ -462,6 +488,7 @@ object
   method status = Stub.result_status res
   method error = Stub.result_error res
   method ntuples = ntuples
+  method nparams = Lazy.force nparams
   method nfields = nfields
   method binary_tuples = binary_tuples
   method fname field = check_field field; Stub.fname res field
@@ -478,6 +505,10 @@ object
     check_field field;
     ftype_of_oid (Stub.ftype res field)
 
+  method paramtype field =
+    check_param field;
+    ftype_of_oid (Stub.paramtype res field)
+
   method fmod field = check_field field; Stub.fmod res field
   method fsize field = check_field field; Stub.fsize res field
 
@@ -629,19 +660,26 @@ object(self)
   method empty_result status =
     check_null (); new result (Stub.make_empty_res conn status)
 
-  method exec ?(expect = []) query =
+  method exec ?(expect = []) ?(params = [||]) ?(binary_params = [||]) query =
     check_null ();
-    let r = Stub.exec conn query in
+    let r = Stub.exec_params conn query params binary_params in
     if Stub.result_isnull r then signal_error ();
     let res = new result r in
     let stat = res#status in
     if not (expect = []) && not (List.mem stat expect) then
-      raise (Error (Unexpected_status (stat, res#error, expect)));
-    res
+      raise (Error (Unexpected_status (stat, res#error, expect)))
+    else res
 
-  method send_query query =
+  method describe_prepared query =
     check_null ();
-    if Stub.send_query conn query <> 1 then signal_error ()
+    let r = Stub.describe_prepared conn query in
+    if Stub.result_isnull r then signal_error ()
+    else new result r
+
+  method send_query ?(params = [||]) ?(binary_params = [||]) query =
+    check_null ();
+    if Stub.send_query_params conn query params binary_params <> 1 then
+      signal_error ()
 
   method get_result =
     check_null ();
