@@ -24,8 +24,6 @@
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)
 
-(* $Id: postgresql.mli,v 1.5 2006/01/24 21:11:09 mottl Exp $ *)
-
 (** Client-interface to the PostgreSQL database. *)
 
 (** Please learn about more details in the database documentation! *)
@@ -203,6 +201,13 @@ val invalid_oid : oid
 (** [invalid_oid] invalid Oid. *)
 
 
+(** {6 Query parameters} *)
+
+val null : string
+(** [null] can be used as an element of the optional argument [parameters]
+    passed to the [exec] or [send_query] method to indicate a NULL value. *)
+
+
 (** Class type of query results.
 
     Indices of tuples and fields start at 0!
@@ -222,6 +227,12 @@ class type result = object
   method ntuples : int
   (** [#ntuples] @return the number of tuples of a query result. *)
 
+  method nparams : int
+  (** [#nparams] @return the number of parameters of a prepared
+      statement.  This function is only useful when inspecting the result
+      of [#describe_prepared].  For other types of queries it will return
+      zero. *)
+
   method nfields : int
   (** [#nfields] @return the number of fields in a query result. *)
 
@@ -250,6 +261,16 @@ class type result = object
       @raise Error if field out of range.
   *)
 
+  method paramtype : int -> ftype
+  (** [#paramtype n] @return the datatype of the indicated statement
+      parameter.  Parameter numbers start at 0.  This function is
+      only useful when inspecting the result of [#describe_prepared].
+      For other types of queries it will return zero.
+
+      @raise Oid if there was no corresponding ftype for the internal [oid].
+      @raise Error if field out of range.
+  *)
+
   method fmod : int -> int
   (** [#fmod n] @return type-specific modification data of the [n]th field.
 
@@ -488,23 +509,76 @@ object
       @raise Error if there is a connection error.
   *)
 
-  method exec : ?expect : result_status list -> string -> result
-  (** [exec ?expect query] synchronous execution of query or command
-      [query]. The result status will be checked against all elements
-      in [expect]. If [expect] is not empty and if there is no match,
-      the exception [Unexpected_status] will be raised.
+  method exec :
+    ?expect : result_status list -> ?params : string array ->
+    ?binary_params : bool array -> string -> result
+  (** [exec ?expect ?params ?binary_params query] synchronous execution
+      of query or command [query].  The result status will be checked
+      against all elements in [expect].  If [expect] is not empty and if
+      there is no match, the exception [Unexpected_status] will be raised.
+
+      Additional query parameters can be passed in the [params] array.
+      They must not be escaped and they can be referred to in [query]
+      as $1, $2, ...  The value [null] can be used in the [params]
+      array to denote an SQL NULL. It is possible to specify that some
+      of the query parameters are passed as binary strings using the
+      [binary_params] array.
+
+      If no (or an empty) query parameter is passed, it is possible to
+      emit several commands with a single call.
 
       @return result of query.
 
       @param expect default = []
+      @param params default = [||]
+      @param binary_params default = [||]
 
       @raise Error if there is a connection error.
       @raise Error if there is an unexpected result status.
   *)
 
-  method send_query : string -> unit
-  (** [send_query query] asynchronous execution of query or command
-      [query].
+  method describe_prepared : string -> result
+  (** [#describe_prepared stm_name] submits a request to obtain
+      information about the specified prepared statement, and waits for
+      completion.  {!describe_prepared} allows an application to obtain
+      information about a previously prepared statement.  The [stm_name]
+      parameter can be the empty string ("") to reference the unnamed
+      statement, otherwise it must be the name of an existing prepared
+      statement.  On success, a {!result} with status [Command_ok] is
+      returned.  The methods {!result.nparams} and {!result.paramtype}
+      of the class [result] can be used to obtain information about
+      the parameters of the prepared statement, and the methods
+      {!result.nfields}, {!result.fname} and {!result.ftype} provide
+      information about the result columns (if any) of the statement.
+
+      To prepare a statement use the SQL command PREPARE.
+
+      @param stm_name The name of the previously prepared query
+
+      @raise Error if there is a connection error.
+
+      @see <http://www.postgresql.org/docs/8.3/interactive/sql-prepare.html>
+      PostgreSQL documentation about [PREPARE]
+  *)
+
+  method send_query :
+    ?params : string array -> ?binary_params : bool array
+    -> string -> unit
+  (** [send_query ?params ?binary_params query] asynchronous execution
+      of query or command [query].
+
+      Additional query parameters can be passed in the [params] array.
+      They must not be escaped and they can be referred to in [query]
+      as $1, $2, ...   The value [null] can be used in the [params]
+      array to denote an SQL NULL. It is possible to specify that some
+      of the query parameters are passed as binary strings using the
+      [binary_params] array.
+
+      If no (or an empty) query parameter is passed, it is possible to
+      emit several commands with a single call.
+
+      @param params default = [||]
+      @param binary_params default = [||]
 
       @raise Error if there is a connection error.
   *)
