@@ -24,7 +24,18 @@
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
-/* $Id: postgresql_stubs.c,v 1.15 2006/01/27 22:59:34 mottl Exp $ */
+#if __GNUC__ >= 3
+# define inline inline __attribute__ ((always_inline))
+# define __unused __attribute__ ((unused))
+#else
+# define __unused
+# define inline
+#endif
+
+#if PG_OCAML_MAJOR_VERSION > 8 \
+    || ( PG_OCAML_MAJOR_VERSION >= 8 && PG_OCAML_MINOR_VERSION >= 2)
+# define PG_OCAML_8_2
+#endif
 
 #include <string.h>
 
@@ -99,7 +110,6 @@
 #define OPAQUEOID               2282
 #define ANYELEMENTOID           2283
 
-static value v_empty_string = 0;
 static value v_None = Val_int(0);
 
 static inline value make_some(value v)
@@ -110,14 +120,17 @@ static inline value make_some(value v)
   CAMLreturn(v_res);
 }
 
-/* Cache for exceptions */
+/* Cache for OCaml-values */
+static value v_empty_string = Val_unit;
 static value *v_exc_Oid = NULL;  /* Exception [Oid] */
+static value *v_null_param = NULL;
 
-CAMLprim value PQocaml_init(value v_unit)
+CAMLprim value PQocaml_init(value __unused v_unit)
 {
-  register_global_root(&v_empty_string);
   v_empty_string = caml_alloc_string(0);
+  caml_register_generational_global_root(&v_empty_string);
   v_exc_Oid = caml_named_value("Postgresql.Oid");
+  v_null_param = caml_named_value("Postgresql.null");
   return Val_unit;
 }
 
@@ -171,7 +184,7 @@ static inline np_callback * np_new(value v_handler)
   c = (np_callback *) caml_stat_alloc(sizeof(np_callback));
   c->v_cb = v_handler;
   c->cnt = 1;
-  register_global_root(&(c->v_cb));
+  caml_register_generational_global_root(&(c->v_cb));
   return c;
 }
 
@@ -182,8 +195,8 @@ static inline void np_decr_refcount(np_callback *c)
   if (c) {
     c->cnt--;
     if (c->cnt == 0) {
-      remove_global_root(&c->v_cb);
-      stat_free(c);
+      caml_remove_generational_global_root(&c->v_cb);
+      caml_stat_free(c);
     }
   }
 }
@@ -226,7 +239,7 @@ CAMLprim value PQconnectdb_stub(value v_conn_info)
   PGconn *conn;
   value v_conn;
 
-  int len = string_length(v_conn_info) + 1;
+  int len = caml_string_length(v_conn_info) + 1;
   char *conn_info = caml_stat_alloc(len);
   memcpy(conn_info, String_val(v_conn_info), len);
 
@@ -261,7 +274,7 @@ CAMLprim value PQreset_stub(value v_conn)
   CAMLreturn(Val_unit);
 }
 
-CAMLprim value PQconndefaults_stub(value v_unit)
+CAMLprim value PQconndefaults_stub(value __unused v_unit)
 {
   CAMLparam0();
   CAMLlocal1(v_res);
@@ -278,18 +291,18 @@ CAMLprim value PQconndefaults_stub(value v_unit)
     for (j = 0; j < 7; j++) { Field(v_el, j) = v_None; };
     Store_field(v_res, i, v_el);
     Field(v_el, 0) = caml_copy_string(cios->keyword);
-    modify(&Field(v_el, 1), caml_copy_string(cios->envvar));
+    caml_modify(&Field(v_el, 1), caml_copy_string(cios->envvar));
     if (cios->compiled) {
       value v_Some = make_some(caml_copy_string(cios->compiled));
-      modify(&Field(v_el, 2), v_Some);
+      caml_modify(&Field(v_el, 2), v_Some);
     };
     if (cios->val) {
       value v_Some = make_some(caml_copy_string(cios->val));
-      modify(&Field(v_el, 3), v_Some);
+      caml_modify(&Field(v_el, 3), v_Some);
     };
-    modify(&Field(v_el, 4), caml_copy_string(cios->label));
-    modify(&Field(v_el, 5), caml_copy_string(cios->dispchar));
-    modify(&Field(v_el, 6), Val_int(cios->dispsize));
+    caml_modify(&Field(v_el, 4), caml_copy_string(cios->label));
+    caml_modify(&Field(v_el, 5), caml_copy_string(cios->dispchar));
+    caml_modify(&Field(v_el, 6), Val_int(cios->dispsize));
   };
 
   CAMLreturn(v_res);
@@ -399,20 +412,134 @@ static inline value alloc_result(PGresult *res, np_callback *cb)
   return v_res;
 }
 
-CAMLprim value PQexec_stub(value v_conn, value v_query)
+static inline void copy_binary_params(
+  value v_params, value v_binary_params, int nparams,
+  int **res_formats, int **res_lengths)
+{
+  int i, nbinary, *lengths, *formats;
+  nbinary = Wosize_val(v_binary_params);
+  if (nbinary == 0 || nparams == 0) {
+    *res_formats = NULL;
+    *res_lengths = NULL;
+    return;
+  }
+  lengths = caml_stat_alloc(nparams * sizeof(int));
+  formats = caml_stat_alloc(nparams * sizeof(int));
+  for (i = 0; i < nparams; i++) {
+    formats[i] = 0;
+    lengths[i] = 0;
+  }
+  if (nbinary > nparams) nbinary = nparams;
+  for (i = 0; i < nbinary; i++) {
+    if (Bool_val(Field(v_binary_params, i))) {
+      formats[i] = 1;
+      lengths[i] = caml_string_length(Field(v_params, i));
+    }
+  }
+  *res_formats = formats;
+  *res_lengths = lengths;
+}
+
+static inline void free_binary_params(int *formats, int *lengths)
+{
+  if (formats != NULL) free(formats);
+  if (lengths != NULL) free(lengths);
+}
+
+static inline const char * const * copy_params(value v_params, int nparams)
+{
+  char **params;
+  int i;
+  if (nparams == 0) return NULL;
+  params = caml_stat_alloc(nparams * sizeof(char *));
+  for (i = 0; i < nparams; i++) {
+    value v_param = Field(v_params, i);
+    if (v_param == *v_null_param) params[i] = NULL;
+    else {
+      int param_len = caml_string_length(v_param) + 1;
+      params[i] = caml_stat_alloc(param_len);
+      memcpy(params[i], String_val(v_param), param_len);
+    }
+  }
+  return (const char * const *) params;
+}
+
+static inline void free_params(const char * const *params, int nparams)
+{
+  int i;
+  if (nparams == 0) return;
+  for (i = 0; i < nparams; i++) caml_stat_free((char *) params[i]);
+  free((char **) params);
+}
+
+static inline const char * const * copy_params_shallow(
+  value v_params, int nparams)
+{
+  char **params;
+  int i;
+  if (nparams == 0) return NULL;
+  params = caml_stat_alloc(nparams * sizeof(char *));
+  for (i = 0; i < nparams; i++) {
+    value v_param = Field(v_params, i);
+    params[i] = (v_param == *v_null_param) ? NULL : String_val(v_param);
+  }
+  return (const char * const *) params;
+}
+
+static inline void free_params_shallow(const char * const *params, int nparams)
+{
+  if (nparams == 0) return;
+  free((char **) params);
+}
+
+CAMLprim value PQexecParams_stub(
+  value v_conn, value v_query, value v_params, value v_binary_params)
 {
   CAMLparam1(v_conn);
   PGconn *conn = get_conn(v_conn);
   np_callback *np_cb = get_conn_cb(v_conn);
   PGresult *res;
-  int len = string_length(v_query) + 1;
+  int len = caml_string_length(v_query) + 1;
+  char *query = caml_stat_alloc(len);
+  int nparams = Wosize_val(v_params);
+  const char * const *params = copy_params(v_params, nparams);
+  int *formats, *lengths;
+  copy_binary_params(v_params, v_binary_params, nparams, &formats, &lengths);
+  memcpy(query, String_val(v_query), len);
+  caml_enter_blocking_section();
+    res =
+      (nparams == 0)
+        ? PQexec(conn, query)
+        : PQexecParams(conn, query, nparams, NULL, params, lengths, formats, 0);
+    free(query);
+    free_params(params, nparams);
+    free_binary_params(formats, lengths);
+  caml_leave_blocking_section();
+  CAMLreturn(alloc_result(res, np_cb));
+}
+
+#ifdef PG_OCAML_8_2
+CAMLprim value PQdescribePrepared_stub(value v_conn, value v_query)
+{
+  CAMLparam1(v_conn);
+  PGconn *conn = get_conn(v_conn);
+  np_callback *np_cb = get_conn_cb(v_conn);
+  PGresult *res;
+  int len = caml_string_length(v_query) + 1;
   char *query = caml_stat_alloc(len);
   memcpy(query, String_val(v_query), len);
   caml_enter_blocking_section();
-    res = PQexec(conn, query);
+    res = PQdescribePrepared(conn, query);
     free(query);
   caml_leave_blocking_section();
   CAMLreturn(alloc_result(res, np_cb));
+#else
+CAMLprim value
+PQdescribePrepared_stub(value __unused v_conn, value __unused v_query)
+{
+  caml_failwith("Postgresql.describe_prepared: not supported");
+  return Val_unit;
+#endif
 }
 
 noalloc_res_info(PQresultStatus, Val_int)
@@ -428,6 +555,16 @@ noalloc_res_info(PQnfields, Val_int)
 noalloc_res_info(PQbinaryTuples, Val_bool)
 fieldnum_info(PQfname, make_string)
 
+#ifdef PG_OCAML_8_2
+noalloc_res_info(PQnparams, Val_int)
+#else
+CAMLprim value PQnparams_stub(value __unused v_res)
+{
+  caml_failwith("Postgresql.nparams: not supported");
+  return Val_unit;
+}
+#endif
+
 CAMLprim value PQfnumber_stub(value v_res, value v_field_name)
 {
   return Val_int(PQfnumber(get_res(v_res), String_val(v_field_name)));
@@ -438,6 +575,18 @@ noalloc_fieldnum_info(PQftype, Val_int)
 noalloc_fieldnum_info(PQfsize, Val_int)
 noalloc_fieldnum_info(PQfmod, Val_int)
 
+#ifdef PG_OCAML_8_2
+noalloc_fieldnum_info(PQparamtype, Val_int)
+#else
+CAMLprim value
+PQparamtype_stub(value __unused v_res, value __unused v_field_num)
+{
+  caml_failwith("Postgresql.paramtype: not supported");
+  return Val_unit;
+}
+#endif
+
+
 CAMLprim value PQgetvalue_stub(value v_res, value v_tup_num, value v_field_num)
 {
   CAMLparam1(v_res);
@@ -482,9 +631,22 @@ CAMLprim value PQsetnonblocking_stub(value v_conn, value v_arg)
 
 noalloc_conn_info(PQisnonblocking, Val_bool)
 
-CAMLprim value PQsendQuery_stub(value v_conn, value v_query)
+CAMLprim value PQsendQueryParams_stub(
+  value v_conn, value v_query, value v_params, value v_binary_params)
 {
-  return Val_int(PQsendQuery(get_conn(v_conn), String_val(v_query)));
+  PGconn *conn = get_conn(v_conn);
+  const char *query = String_val(v_query);
+  int nparams = Wosize_val(v_params);
+  const char * const *params = copy_params_shallow(v_params, nparams);
+  int *lengths, *formats, res;
+  copy_binary_params(v_params, v_binary_params, nparams, &formats, &lengths);
+  res =
+    (nparams == 0)
+      ? PQsendQuery(conn, query)
+      : PQsendQueryParams(conn, query, nparams, NULL, params, lengths, formats, 0);
+  free_params_shallow(params, nparams);
+  free_binary_params(formats, lengths);
+  return Val_int(res);
 }
 
 CAMLprim value PQgetResult_stub(value v_conn)
@@ -531,16 +693,15 @@ CAMLprim value PQescapeByteaConn_stub(
 CAMLprim value PQunescapeBytea_stub(value v_from)
 {
   size_t len;
+  value v_res;
   char *buf =
     (char *) PQunescapeBytea((unsigned char *) String_val(v_from), &len);
   if (buf == NULL)
     caml_failwith("Postgresql.unescape_bytea: illegal bytea string");
-  else {
-    value v_res = caml_alloc_string(len);
-    memcpy(String_val(v_res), buf, len);
-    PQfreemem(buf);
-    return v_res;
-  }
+  v_res = caml_alloc_string(len);
+  memcpy(String_val(v_res), buf, len);
+  PQfreemem(buf);
+  return v_res;
 }
 
 
@@ -595,7 +756,7 @@ CAMLprim value PQputline_stub(value v_conn, value v_buf)
   CAMLparam1(v_conn);
   PGconn *conn = get_conn(v_conn);
   value v_res;
-  int len = string_length(v_buf) + 1;
+  int len = caml_string_length(v_buf) + 1;
   char *buf = caml_stat_alloc(len);
   memcpy(buf, String_val(v_buf), len);
   caml_enter_blocking_section();
@@ -714,12 +875,12 @@ CAMLprim value lo_lseek_stub(
   PGconn *conn = get_conn(v_conn);
   value v_res;
   int whence;
-  switch (Int_val(v_whence)) {
-    case 0 : whence = SEEK_SET; break;
-    case 1 : whence = SEEK_CUR; break;
-    default : whence = SEEK_END; break;
-  }
   caml_enter_blocking_section();
+    switch (Int_val(v_whence)) {
+      case 0 : whence = SEEK_SET; break;
+      case 1 : whence = SEEK_CUR; break;
+      default : whence = SEEK_END; break;
+    }
     v_res = Val_int(lo_lseek(conn, Int_val(v_fd), Int_val(v_pos), whence));
   caml_leave_blocking_section();
   CAMLreturn(v_res);
@@ -763,7 +924,7 @@ CAMLprim value lo_import_stub(value v_conn, value v_fname)
   CAMLparam1(v_conn);
   PGconn *conn = get_conn(v_conn);
   value v_res;
-  int len = string_length(v_fname) + 1;
+  int len = caml_string_length(v_fname) + 1;
   char *fname = caml_stat_alloc(len);
   memcpy(fname, String_val(v_fname), len);
   caml_enter_blocking_section();
@@ -778,7 +939,7 @@ CAMLprim value lo_export_stub(value v_conn, value v_oid, value v_fname)
   CAMLparam1(v_conn);
   PGconn *conn = get_conn(v_conn);
   value v_res;
-  int len = string_length(v_fname) + 1;
+  int len = caml_string_length(v_fname) + 1;
   char *fname = caml_stat_alloc(len);
   memcpy(fname, String_val(v_fname), len);
   caml_enter_blocking_section();
